/**
 * @file
 *
 * @ingroup demo
 *
 * @brief Directory iteration implementatin.
 */

/*
 * Copyright (c) 2009, 2010 embedded brains GmbH.  All rights reserved.
 *
 *  embedded brains GmbH
 *  Obere Lagerstr. 30
 *  82178 Puchheim
 *  Germany
 *  <rtems@embedded-brains.de>
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>

#include "media.h"

rtems_status_code rtems_directory_iterate(
  const char *dir_path,
  rtems_per_directory_routine routine,
  void *routine_arg
)
{
  rtems_status_code rsc = RTEMS_SUCCESSFUL;
  int rv = 0;
  size_t const path_len = strlen(dir_path);
  DIR *const dir = opendir(dir_path);
  struct dirent *de = NULL;
  size_t buf_len = 0;
  char *buf = NULL;

  if (dir == NULL) {
    return RTEMS_INVALID_ID;
  }

  while ((de = readdir(dir)) != NULL) {
    size_t const new_buf_len = path_len + strlen(de->d_name) + 2;
    struct stat st;

    if (new_buf_len > buf_len) {
      free(buf);
      buf = malloc(new_buf_len);
      if (buf == NULL) {
        rsc = RTEMS_NO_MEMORY;

        break;
      }

      strcpy(buf, dir_path);
      buf [path_len] = '/';
    }

    strcpy(buf + path_len + 1, de->d_name);

    rv = lstat(buf, &st);
    if (rv != 0) {
      rsc = RTEMS_IO_ERROR;

      break;
    }

    rsc = (*routine)(buf, &st, routine_arg);
    if (rsc != RTEMS_SUCCESSFUL) {
      break;
    }
  }

  free(buf);
  closedir(dir);

  return rsc;
}
