/**
 * @file
 *
 * @ingroup rtems_media
 *
 * @brief Media implementation.
 */

/*
 * Copyright (c) 2009, 2010 embedded brains GmbH.  All rights reserved.
 *
 *  embedded brains GmbH
 *  Obere Lagerstr. 30
 *  82178 Puchheim
 *  Germany
 *  <rtems@embedded-brains.de>
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

#include "media.h"

static const char *const rtems_media_event_assoc_table [] = {
  [RTEMS_MEDIA_EVENT_DISK_ATTACH] = "DISK ATTACH",
  [RTEMS_MEDIA_EVENT_DISK_DETACH] = "DISK DETACH",
  [RTEMS_MEDIA_EVENT_MOUNT] = "MOUNT",
  [RTEMS_MEDIA_EVENT_UNMOUNT] = "UNMOUNT",
  [RTEMS_MEDIA_EVENT_PARTITION_INQUIRY] = "PARTITION INQUIRY",
  [RTEMS_MEDIA_EVENT_PARTITION_ATTACH] = "PARTITION ATTACH",
  [RTEMS_MEDIA_EVENT_PARTITION_DETACH] = "PARTITION DETACH"
};

static const char *const rtems_media_state_assoc_table [] = {
  [RTEMS_MEDIA_STATE_INQUIRY] = "INQUIRY",
  [RTEMS_MEDIA_STATE_ABORTED] = "ABORTED",
  [RTEMS_MEDIA_STATE_READY] = "SUCCESS",
  [RTEMS_MEDIA_STATE_FAILED] = "FAILED",
  [RTEMS_MEDIA_STATE_SUCCESS] = "SUCCESS"
};

#define TC(table) (sizeof(table) / sizeof(table [0]))

const char *rtems_media_event_assoc(rtems_media_event event)
{
  if ((size_t) event < TC(rtems_media_event_assoc_table)) {
    return rtems_media_event_assoc_table [event];
  } else {
    return "INVALID";
  } 
}

const char *rtems_media_state_assoc(rtems_media_state state)
{
  if ((size_t) state < TC(rtems_media_state_assoc_table)) {
    return rtems_media_state_assoc_table [state];
  } else {
    return "INVALID";
  }
}
