#include <rtems/freebsd/machine/rtems-bsd-config.h>

#include <bsp.h>

#ifdef __GEN83xx_BSP_h

#include <bsp/irq.h>
#include <mpc83xx/mpc83xx.h>

#include <rtems/freebsd/sys/cdefs.h>
#include <rtems/freebsd/sys/stdint.h>
#include <rtems/freebsd/sys/stddef.h>
#include <rtems/freebsd/sys/param.h>
#include <rtems/freebsd/sys/queue.h>
#include <rtems/freebsd/sys/types.h>
#include <rtems/freebsd/sys/systm.h>
#include <rtems/freebsd/sys/kernel.h>
#include <rtems/freebsd/sys/bus.h>
#include <rtems/freebsd/sys/linker_set.h>
#include <rtems/freebsd/sys/module.h>
#include <rtems/freebsd/sys/lock.h>
#include <rtems/freebsd/sys/mutex.h>
#include <rtems/freebsd/sys/condvar.h>
#include <rtems/freebsd/sys/sysctl.h>
#include <rtems/freebsd/sys/sx.h>
#include <rtems/freebsd/sys/unistd.h>
#include <rtems/freebsd/sys/callout.h>
#include <rtems/freebsd/sys/malloc.h>
#include <rtems/freebsd/sys/priv.h>

#include <rtems/freebsd/dev/usb/usb.h>
#include <rtems/freebsd/dev/usb/usbdi.h>

#include <rtems/freebsd/dev/usb/usb_core.h>
#include <rtems/freebsd/dev/usb/usb_busdma.h>
#include <rtems/freebsd/dev/usb/usb_process.h>
#include <rtems/freebsd/dev/usb/usb_util.h>

#include <rtems/freebsd/dev/usb/usb_controller.h>
#include <rtems/freebsd/dev/usb/usb_bus.h>
#include <rtems/freebsd/dev/usb/controller/ehci.h>

static device_probe_t ehci_mpc83xx_probe;
static device_attach_t ehci_mpc83xx_attach;
static device_detach_t ehci_mpc83xx_detach;
static device_suspend_t ehci_mpc83xx_suspend;
static device_resume_t ehci_mpc83xx_resume;

static int
ehci_mpc83xx_suspend(device_t self)
{
	ehci_softc_t *e = device_get_softc(self);
	int eno = bus_generic_suspend(self);

	if (eno != 0) {
		return (eno);
	}

	ehci_suspend(e);

	return (0);
}

static int
ehci_mpc83xx_resume(device_t self)
{
	ehci_softc_t *e = device_get_softc(self);

	ehci_resume(e);

	bus_generic_resume(self);

	return (0);
}


static int
ehci_mpc83xx_probe(device_t self)
{
	device_set_desc(self, "MPC83XX EHCI Controller");

	return (0);
}

#define CONTROL_ULPI_INT_EN (1U << 0)
#define CONTROL_WU_INT_EN (1U << 1)
#define CONTROL_USB_EN (1U << 2)
#define CONTROL_LSF_EN (1U << 3)
#define CONTROL_KEEP_OTG_ON (1U << 4)
#define CONTROL_OTG_PORT (1U << 5)
#define CONTROL_REFSEL(x) ((x) << 6)
#define CONTROL_PLL_RESET (1U << 8)
#define CONTROL_UTMI_PHY_EN (1U << 9)
#define CONTROL_PHY_CLOCK_SEL (1U << 10)
#define CONTROL_CLKIN_SEL(x) ((x) << 11)
#define CONTROL_WU_INT (1U << 16)
#define CONTROL_PHY_CLK_VALID (1U << 17)

static void
ehci_mpc83xx_phy_init(void)
{
	volatile uint32_t *control = &mpc83xx.usb_dr.control;

	*control = CONTROL_PLL_RESET | CONTROL_REFSEL(0x2U);
	*control = CONTROL_UTMI_PHY_EN | CONTROL_REFSEL(0x2U);

	while ((*control & CONTROL_PHY_CLK_VALID) == 0) {
		/* Wait for PLL */
	}
}

#define USBMODE_SDIS (1U << 4)
#define USBMODE_SLOM (1U << 3)
#define USBMODE_CM_IDLE (0x0U << 0)
#define USBMODE_CM_DEVICE (0x2U << 0)
#define USBMODE_CM_HOST (0x3U << 0)

#define PORTSC_PTS_UTMI (0x0U << 0)
#define PORTSC_PTS_ULPI (0x2U << 0)

#define SNOOP_SIZE_2GB (0x1eU << 0)

static void
ehci_mpc83xx_host_init(void)
{
	mpc83xx.usb_dr.snoop1 = 0x0U | SNOOP_SIZE_2GB;
	mpc83xx.usb_dr.snoop2 = 0x80000000U | SNOOP_SIZE_2GB;
	mpc83xx.usb_dr.control = CONTROL_UTMI_PHY_EN | CONTROL_REFSEL(0x2U) | CONTROL_USB_EN;
	mpc83xx.usb_dr.portsc1 = PORTSC_PTS_UTMI;
#if 0
	mpc83xx.usb_dr.pri_ctrl = 0xcU;
	mpc83xx.usb_dr.age_cnt_thresh = 0x40U;
	mpc83xx.usb_dr.si_ctrl = 0x1U;
#endif
	mpc83xx.usb_dr.usbmode = htole32(USBMODE_CM_HOST);
}

static int
ehci_mpc83xx_attach(device_t self)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;
	ehci_softc_t *e = device_get_softc(self);
	usb_error_t ue = USB_ERR_NORMAL_COMPLETION;
	int eno = 0;

	memset(e, 0, sizeof(*e));

	/* Initialize some bus fields */
	e->sc_bus.parent = self;
	e->sc_bus.devices = e->sc_devices;
	e->sc_bus.devices_max = EHCI_MAX_DEVICES;
	e->sc_bus.usbrev = USB_REV_2_0;

	/* Get all DMA memory */
	if (usb_bus_mem_alloc_all(&e->sc_bus, USB_GET_DMA_TAG(self), &ehci_iterate_hw_softc)) {
		return (ENOMEM);
	}

	/* Child device */
	e->sc_bus.bdev = device_add_child(self, "usbus", -1);
	if (e->sc_bus.bdev == NULL) {
		device_printf(self, "Could not add USB device\n");
		goto error;
	}
	device_set_ivars(e->sc_bus.bdev, &e->sc_bus);
	device_set_desc(e->sc_bus.bdev, "MPC83XX EHCI Bus");
	snprintf(e->sc_vendor, sizeof(e->sc_vendor), "Freescale");

	/* Register space */
	e->sc_io_tag = 0x0U;
	e->sc_io_hdl = (unsigned) &mpc83xx.usb_dr.caplength;
	e->sc_io_size = 0x88U;

	ehci_mpc83xx_phy_init();

	ehci_mpc83xx_host_init();

	/* Install interrupt handler */
	sc = rtems_interrupt_server_handler_install(
		RTEMS_ID_NONE,
		BSP_IPIC_IRQ_USB_DR,
		"USB",
		RTEMS_INTERRUPT_UNIQUE,
		(rtems_interrupt_handler) ehci_interrupt,
		e
	);
	BSD_ASSERT_SC(sc);

	e->sc_flags |= EHCI_SCFLG_SETMODE | EHCI_SCFLG_NORESTERM | EHCI_SCFLG_TT;

	/* EHCI intitialization */
	ue = ehci_init(e);
	if (ue != USB_ERR_NORMAL_COMPLETION) {
		goto error;
	}

	/* Probe and attach child */
	eno = device_probe_and_attach(e->sc_bus.bdev);
	if (eno != 0) {
		goto error;
	}

	return (0);

error:
	ehci_mpc83xx_detach(self);

	return (ENXIO);
}

static int
ehci_mpc83xx_detach(device_t self)
{
	BSD_PRINTF("TODO\n");

	return (0);
}

static device_method_t ehci_methods [] = {
	/* Device interface */
	DEVMETHOD(device_probe, ehci_mpc83xx_probe),
	DEVMETHOD(device_attach, ehci_mpc83xx_attach),
	DEVMETHOD(device_detach, ehci_mpc83xx_detach),
	DEVMETHOD(device_suspend, ehci_mpc83xx_suspend),
	DEVMETHOD(device_resume, ehci_mpc83xx_resume),
	DEVMETHOD(device_shutdown, bus_generic_shutdown),

	/* Bus interface */
	DEVMETHOD(bus_print_child, bus_generic_print_child),

	{0, 0}
};

static driver_t ehci_driver = {
	.name = "ehci",
	.methods = ehci_methods,
	.size = sizeof(struct ehci_softc)
};

static devclass_t ehci_devclass;

DRIVER_MODULE(ehci, nexus, ehci_driver, ehci_devclass, 0, 0);
MODULE_DEPEND(ehci, usb, 1, 1, 1);

#endif /* __GEN83xx_BSP_h */
