#include <rtems/freebsd/machine/rtems-bsd-config.h>

#include <bsp.h>

#ifdef LIBBSP_ARM_LPC32XX_BSP_H

#include <bsp/irq.h>
#include <bsp/lpc32xx.h>

#include <rtems/freebsd/sys/cdefs.h>
#include <rtems/freebsd/sys/stdint.h>
#include <rtems/freebsd/sys/stddef.h>
#include <rtems/freebsd/sys/param.h>
#include <rtems/freebsd/sys/queue.h>
#include <rtems/freebsd/sys/types.h>
#include <rtems/freebsd/sys/systm.h>
#include <rtems/freebsd/sys/kernel.h>
#include <rtems/freebsd/sys/bus.h>
#include <rtems/freebsd/sys/linker_set.h>
#include <rtems/freebsd/sys/module.h>
#include <rtems/freebsd/sys/lock.h>
#include <rtems/freebsd/sys/mutex.h>
#include <rtems/freebsd/sys/condvar.h>
#include <rtems/freebsd/sys/sysctl.h>
#include <rtems/freebsd/sys/sx.h>
#include <rtems/freebsd/sys/unistd.h>
#include <rtems/freebsd/sys/callout.h>
#include <rtems/freebsd/sys/malloc.h>
#include <rtems/freebsd/sys/priv.h>

#include <rtems/freebsd/dev/usb/usb.h>
#include <rtems/freebsd/dev/usb/usbdi.h>

#include <rtems/freebsd/dev/usb/usb_core.h>
#include <rtems/freebsd/dev/usb/usb_busdma.h>
#include <rtems/freebsd/dev/usb/usb_process.h>
#include <rtems/freebsd/dev/usb/usb_util.h>

#include <rtems/freebsd/dev/usb/usb_controller.h>
#include <rtems/freebsd/dev/usb/usb_bus.h>
#include <rtems/freebsd/dev/usb/controller/ohci.h>

#define USB_CTRL_SLAVE_HCLK_EN (1U << 24)
#define USB_CTRL_I2C_EN (1U << 23)
#define USB_CTRL_DEV_NEED_CLK_EN (1U << 22)
#define USB_CTRL_HOST_NEED_CLK_EN (1U << 21)
#define USB_CTRL_PC_MASK (0x3U << 19)
#define USB_CTRL_PC_PULL_UP (0x0U << 19)
#define USB_CTRL_PC_BUS_KEEPER (0x1U << 19)
#define USB_CTRL_PC_NONE (0x2U << 19)
#define USB_CTRL_PC_PULL_DOWN (0x3U << 19)
#define USB_CTRL_CLKEN2 (1U << 18)
#define USB_CTRL_CLKEN1 (1U << 17)
#define USB_CTRL_POWER_UP (1U << 16)
#define USB_CTRL_BYPASS (1U << 15)
#define USB_CTRL_DIRECT (1U << 14)
#define USB_CTRL_FEEDBACK (1U << 13)
#define USB_CTRL_P_SHIFT 11
#define USB_CTRL_P_MASK (0x3U << USB_CTRL_P_SHIFT)
#define USB_CTRL_P_1 (0x0U << USB_CTRL_P_SHIFT)
#define USB_CTRL_P_2 (0x1U << USB_CTRL_P_SHIFT)
#define USB_CTRL_P_4 (0x2U << USB_CTRL_P_SHIFT)
#define USB_CTRL_P_8 (0x3U << USB_CTRL_P_SHIFT)
#define USB_CTRL_N_SHIFT 9
#define USB_CTRL_N_MASK (0x3U << USB_CTRL_N_SHIFT)
#define USB_CTRL_N_1 (0x0U << USB_CTRL_N_SHIFT)
#define USB_CTRL_N_2 (0x1U << USB_CTRL_N_SHIFT)
#define USB_CTRL_N_3 (0x2U << USB_CTRL_N_SHIFT)
#define USB_CTRL_N_4 (0x3U << USB_CTRL_N_SHIFT)
#define USB_CTRL_M_SHIFT 1
#define USB_CTRL_M_MASK (0xffU << USB_CTRL_M_SHIFT)
#define USB_CTRL_PLL_LOCK (1U << 0)

#define I2C_CTL_SRST (1U << 8)

#define I2C_TX_DATA_MASK 0xffU
#define I2C_TX_ADDR_SHIFT 1
#define I2C_TX_ADDR_MASK 0x7fU
#define I2C_TX_READ (1U << 0)
#define I2C_TX_START (1U << 8)
#define I2C_TX_STOP (1U << 9)

#define I2C_STS_TDI (1U << 0)
#define I2C_STS_AFI (1U << 1)
#define I2C_STS_NAI (1U << 2)
#define I2C_STS_RFE (1U << 9)

#define ISP1301_VENDOR_ID_LOW 0x0U
#define ISP1301_VENDOR_ID_HIGH 0x1U
#define ISP1301_PRODUCT_ID_LOW 0x2U
#define ISP1301_PRODUCT_ID_HIGH 0x3U
#define ISP1301_VERSION_ID_LOW 0x14U
#define ISP1301_VERSION_ID_HIGH 0x15U
#define ISP1301_MODE_CONTROL_1 0x4U
#define ISP1301_MODE_CONTROL_1_SET 0x4U
#define ISP1301_MODE_CONTROL_1_CLEAR 0x5U
#define ISP1301_MODE_CONTROL_2 0x12U
#define ISP1301_MODE_CONTROL_2_SET 0x12U
#define ISP1301_MODE_CONTROL_2_CLEAR 0x13U
#define ISP1301_OTG_CONTROL 0x6U
#define ISP1301_OTG_CONTROL_SET 0x6U
#define ISP1301_OTG_CONTROL_CLEAR 0x7U
#define ISP1301_OTG_STATUS 0x10U
#define ISP1301_INTERRUPT_SOURCE 0x8U
#define ISP1301_INTERRUPT_LATCH 0xaU
#define ISP1301_INTERRUPT_LATCH_SET 0xaU
#define ISP1301_INTERRUPT_LATCH_CLEAR 0xbU
#define ISP1301_INTERRUPT_ENABLE_LOW 0xcU
#define ISP1301_INTERRUPT_ENABLE_LOW_SET 0xcU
#define ISP1301_INTERRUPT_ENABLE_LOW_CLEAR 0xdU
#define ISP1301_INTERRUPT_ENABLE_HIGH 0xeU
#define ISP1301_INTERRUPT_ENABLE_HIGH_SET 0xeU
#define ISP1301_INTERRUPT_ENABLE_HIGH_CLEAR 0xfU

#define ISP1301_MODE_CONTROL_1_SPEED_REG (1U << 0)
#define ISP1301_MODE_CONTROL_1_SUSPEND_REG (1U << 1)
#define ISP1301_MODE_CONTROL_1_DAT_SE0 (1U << 2)
#define ISP1301_MODE_CONTROL_1_TRANSP_EN (1U << 3)
#define ISP1301_MODE_CONTROL_1_BDIS_ACON_EN (1U << 4)
#define ISP1301_MODE_CONTROL_1_OE_INT_EN (1U << 5)
#define ISP1301_MODE_CONTROL_1_UART_EN (1U << 6)

#define ISP1301_MODE_CONTROL_2_GLOBAL_PWR_DN (1U << 0)
#define ISP1301_MODE_CONTROL_2_SPD_SUSP_CTRL (1U << 1)
#define ISP1301_MODE_CONTROL_2_BI_DI (1U << 2)
#define ISP1301_MODE_CONTROL_2_TRANSP_BDIR0 (1U << 3)
#define ISP1301_MODE_CONTROL_2_TRANSP_BDIR1 (1U << 4)
#define ISP1301_MODE_CONTROL_2_AUDIO_EN (1U << 5)
#define ISP1301_MODE_CONTROL_2_PSW_OE (1U << 6)
#define ISP1301_MODE_CONTROL_2_EN2V7 (1U << 7)

#define ISP1301_OTG_CONTROL_DP_PULLUP (1U << 0)
#define ISP1301_OTG_CONTROL_DM_PULLUP (1U << 1)
#define ISP1301_OTG_CONTROL_DP_PULLDOWN (1U << 2)
#define ISP1301_OTG_CONTROL_DM_PULLDOWN (1U << 3)
#define ISP1301_OTG_CONTROL_ID_PULLDOWN (1U << 4)
#define ISP1301_OTG_CONTROL_VBUS_DRV (1U << 5)
#define ISP1301_OTG_CONTROL_VBUS_DISCHRG (1U << 6)
#define ISP1301_OTG_CONTROL_VBUS_CHRG (1U << 7)

#define ISP1301_ADDR (0x2cU << I2C_TX_ADDR_SHIFT)

static void
i2c_wait_for_receive_fifo_not_empty(void)
{
	while ((LPC32XX_I2C_STS & I2C_STS_RFE) != 0) {
		/* Wait */
	}
}

static uint8_t
isp1301_read(uint8_t reg)
{
	LPC32XX_I2C_CTL = I2C_CTL_SRST;

	LPC32XX_I2C_TX = ISP1301_ADDR | I2C_TX_START;

	LPC32XX_I2C_TX = reg;

	LPC32XX_I2C_TX = ISP1301_ADDR | I2C_TX_READ | I2C_TX_START;

	LPC32XX_I2C_TX = I2C_TX_STOP;

	i2c_wait_for_receive_fifo_not_empty();

	return (uint8_t) LPC32XX_I2C_RX;
}

static void
i2c_wait_for_transaction_done(void)
{
	while ((LPC32XX_I2C_STS & I2C_STS_TDI) == 0) {
		/* Wait */
	}

	LPC32XX_I2C_STS = I2C_STS_TDI;
}

static void
isp1301_write(uint8_t reg, uint8_t val)
{

	LPC32XX_I2C_CTL = I2C_CTL_SRST;

	LPC32XX_I2C_TX = ISP1301_ADDR | I2C_TX_START;

	LPC32XX_I2C_TX = reg;

	LPC32XX_I2C_TX = val | I2C_TX_STOP;

	i2c_wait_for_transaction_done();
}

static void
isp1301_dump(void)
{
	BSD_PRINTF(
		"ISP1301: mc1 %02x, mc2 %02x, otgctrl %02x, otgsts %02x, isrc %02x, iltch %02x, ienl %02x, ienh %02x\n",
		isp1301_read(ISP1301_MODE_CONTROL_1),
		isp1301_read(ISP1301_MODE_CONTROL_2),
		isp1301_read(ISP1301_OTG_CONTROL),
		isp1301_read(ISP1301_OTG_STATUS),
		isp1301_read(ISP1301_INTERRUPT_SOURCE),
		isp1301_read(ISP1301_INTERRUPT_LATCH),
		isp1301_read(ISP1301_INTERRUPT_ENABLE_LOW),
		isp1301_read(ISP1301_INTERRUPT_ENABLE_HIGH)
	);
}

static void
isp1301_configure(void)
{
	BSD_PRINTF(
		"ISP1301: vendor 0x%02x%02x, product 0x%02x%02x, version 0x%02x%02x\n",
		isp1301_read(ISP1301_VENDOR_ID_HIGH), isp1301_read(ISP1301_VENDOR_ID_LOW),
		isp1301_read(ISP1301_PRODUCT_ID_HIGH), isp1301_read(ISP1301_PRODUCT_ID_LOW),
		isp1301_read(ISP1301_VERSION_ID_HIGH), isp1301_read(ISP1301_VERSION_ID_LOW)
	);

	isp1301_write(ISP1301_MODE_CONTROL_1_CLEAR, 0xff);
	isp1301_write(ISP1301_MODE_CONTROL_1_SET, ISP1301_MODE_CONTROL_1_SPEED_REG);
	isp1301_write(ISP1301_MODE_CONTROL_2_CLEAR, 0xff);
	isp1301_write(ISP1301_MODE_CONTROL_2_SET, ISP1301_MODE_CONTROL_2_BI_DI | ISP1301_MODE_CONTROL_2_PSW_OE | ISP1301_MODE_CONTROL_2_SPD_SUSP_CTRL);
	isp1301_write(ISP1301_OTG_CONTROL_CLEAR, 0xff);
	isp1301_write(ISP1301_MODE_CONTROL_1_SET, ISP1301_MODE_CONTROL_1_DAT_SE0);
	isp1301_write(ISP1301_OTG_CONTROL_SET, ISP1301_OTG_CONTROL_DM_PULLDOWN | ISP1301_OTG_CONTROL_DP_PULLDOWN);
	isp1301_write(ISP1301_INTERRUPT_LATCH_CLEAR, 0xff);
	isp1301_write(ISP1301_INTERRUPT_ENABLE_LOW_CLEAR, 0xff);
	isp1301_write(ISP1301_INTERRUPT_ENABLE_HIGH_CLEAR, 0xff);
}

static void
isp1301_vbus_on(void)
{
	isp1301_write(ISP1301_OTG_CONTROL_SET, ISP1301_OTG_CONTROL_VBUS_DRV);
}

static int
ohci_lpc32xx_suspend(device_t self)
{
	ohci_softc_t *e = device_get_softc(self);
	int eno = bus_generic_suspend(self);

	if (eno != 0) {
		return (eno);
	}

	ohci_suspend(e);

	return (0);
}

static int
ohci_lpc32xx_resume(device_t self)
{
	ohci_softc_t *e = device_get_softc(self);

	ohci_resume(e);

	bus_generic_resume(self);

	return (0);
}


static int
ohci_lpc32xx_probe(device_t self)
{
	device_set_desc(self, "LPC32XX OHCI controller");

	return (0);
}

static int
ohci_lpc32xx_detach(device_t self)
{
	ohci_softc_t *e = device_get_softc(self);

	BSD_PRINTF("FIXME\n");

	return (0);
}

static int
ohci_lpc32xx_attach(device_t self)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;
	ohci_softc_t *e = device_get_softc(self);
	usb_error_t ue = USB_ERR_NORMAL_COMPLETION;
	int eno = 0;

	memset(e, 0, sizeof(*e));

	/* Initialize some bus fields */
	e->sc_bus.parent = self;
	e->sc_bus.devices = e->sc_devices;
	e->sc_bus.devices_max = OHCI_MAX_DEVICES;

	/* Get all DMA memory */
	if (usb_bus_mem_alloc_all(&e->sc_bus, USB_GET_DMA_TAG(self), &ohci_iterate_hw_softc)) {
		return (ENOMEM);
	}
	e->sc_dev = self;

	/* Child device */
	e->sc_bus.bdev = device_add_child(self, "usbus", -1);
	if (e->sc_bus.bdev == NULL) {
		device_printf(self, "Could not add USB device\n");
		goto error;
	}
	device_set_ivars(e->sc_bus.bdev, &e->sc_bus);
	device_set_desc(e->sc_bus.bdev, "LPC32XX OHCI bus");
	snprintf(e->sc_vendor, sizeof(e->sc_vendor), "NXP");

	/* Register space */
	e->sc_io_tag = 0U;
	e->sc_io_hdl = LPC32XX_BASE_USB;
	e->sc_io_size = 0x5cU;

	/* Enable USB PLL */
	LPC32XX_USB_DIV = 0xc;
	LPC32XX_USB_CTRL = USB_CTRL_SLAVE_HCLK_EN
		| USB_CTRL_PC_BUS_KEEPER
		| USB_CTRL_CLKEN1
		| USB_CTRL_POWER_UP
		| USB_CTRL_P_2
		| USB_CTRL_N_1
		| (191U << USB_CTRL_M_SHIFT);
	while ((LPC32XX_USB_CTRL & USB_CTRL_PLL_LOCK) == 0) {
		/* Wait */
	}
	LPC32XX_USB_CTRL |= USB_CTRL_CLKEN2;

	/* Enable USB host and AHB clocks */
	LPC32XX_OTG_CLK_CTRL = 0x1c;
	while ((LPC32XX_OTG_CLK_STAT & 0x1c) != 0x1c) {
		/* Wait */
	}

	isp1301_configure();

	LPC32XX_USB_CTRL |= USB_CTRL_HOST_NEED_CLK_EN;

	LPC32XX_OTG_CLK_CTRL = 0x1d;
	while ((LPC32XX_OTG_CLK_STAT & 0x1d) != 0x1d) {
		/* Wait */
	}

	/* Set OTG Status and Control Register */
	LPC32XX_OTG_STAT_CTRL = 0x1;

	isp1301_vbus_on();

	/* Install interrupt handler */
	sc = rtems_interrupt_server_handler_install(
		RTEMS_ID_NONE,
		LPC32XX_IRQ_USB_HOST,
		"USB",
		RTEMS_INTERRUPT_UNIQUE,
		(rtems_interrupt_handler) ohci_interrupt,
		e
	);
	BSD_ASSERT_SC(sc);

	/* OHCI intitialization */
	ue = ohci_init(e);
	if (ue != USB_ERR_NORMAL_COMPLETION) {
		goto error;
	}

	/* Probe and attach child */
	eno = device_probe_and_attach(e->sc_bus.bdev);
	if (eno != 0) {
		goto error;
	}

	return (0);

error:
	ohci_lpc32xx_detach(self);
	return (ENXIO);
}

static device_method_t ohci_methods [] = {
	/* Device interface */
	DEVMETHOD(device_probe, ohci_lpc32xx_probe),
	DEVMETHOD(device_attach, ohci_lpc32xx_attach),
	DEVMETHOD(device_detach, ohci_lpc32xx_detach),
	DEVMETHOD(device_suspend, ohci_lpc32xx_suspend),
	DEVMETHOD(device_resume, ohci_lpc32xx_resume),
	DEVMETHOD(device_shutdown, bus_generic_shutdown),

	/* Bus interface */
	DEVMETHOD(bus_print_child, bus_generic_print_child),

	{0, 0}
};

static driver_t ohci_driver = {
	.name = "ohci",
	.methods = ohci_methods,
	.size = sizeof(struct ohci_softc)
};

static devclass_t ohci_devclass;

DRIVER_MODULE(ohci, nexus, ohci_driver, ohci_devclass, 0, 0);
MODULE_DEPEND(ohci, usb, 1, 1, 1);

#endif /* LIBBSP_ARM_LPC32XX_BSP_H */
