/**
 * @file
 *
 * @ingroup rtems_bsd_rtems
 *
 * @brief TODO.
 */

/*
 * Copyright (c) 2009, 2010 embedded brains GmbH.  All rights reserved.
 *
 *  embedded brains GmbH
 *  Obere Lagerstr. 30
 *  82178 Puchheim
 *  Germany
 *  <rtems@embedded-brains.de>
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

#include <rtems/freebsd/machine/rtems-bsd-config.h>

#include <rtems/irq-extension.h>

#include <rtems/freebsd/bsd.h>

#include <rtems/freebsd/sys/param.h>
#include <rtems/freebsd/sys/types.h>
#include <rtems/freebsd/sys/systm.h>
#include <rtems/freebsd/sys/kernel.h>
#include <rtems/freebsd/sys/lock.h>
#include <rtems/freebsd/sys/mutex.h>
#include <rtems/freebsd/sys/proc.h>

/* In FreeBSD this is a local function */
void mi_startup(void);

int hz;

int
pause(const char *wmesg, int timo)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	/* FIXME: Integer conversion */
	sc = rtems_task_wake_after((rtems_interval) timo);
	BSD_ASSERT_SC(sc);

	return EWOULDBLOCK;
}

void
DELAY(int usec)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	/* FIXME: Integer conversion */
	rtems_interval ticks =
		((rtems_interval) usec * (rtems_interval) hz) / 1000000;

	if (ticks == 0) {
		ticks = 1;
	}

	sc = rtems_task_wake_after(ticks);
	BSD_ASSERT_SC(sc);
}

static void
rtems_bsd_suspend_all_threads(void)
{
	rtems_chain_control *chain = &rtems_bsd_thread_chain;
	rtems_chain_node *node = rtems_chain_first(chain);
	rtems_id self = rtems_task_self();

	while (!rtems_chain_is_tail(chain, node)) {
		struct thread *td = (struct thread *) node;

		if (td->td_id != self && td->td_id != RTEMS_SELF) {
			rtems_task_suspend(td->td_id);
		}

		node = rtems_chain_next(node);
	}

	rtems_task_suspend(RTEMS_SELF);
}

void
panic(const char *fmt, ...)
{
	va_list ap;

	printf("*** BSD PANIC *** ");

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	printf("\n");

	rtems_bsd_suspend_all_threads();

	/* FIXME */
	rtems_fatal_error_occurred(0xdeadbeef);
}

rtems_status_code
rtems_bsd_initialize(void)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	hz = (int) rtems_clock_get_ticks_per_second();

	sc = rtems_interrupt_server_initialize(
		BSD_TASK_PRIORITY_INTERRUPT,
		BSD_MINIMUM_TASK_STACK_SIZE,
		RTEMS_DEFAULT_MODES,
		RTEMS_DEFAULT_ATTRIBUTES,
		NULL
	);
	if (sc != RTEMS_SUCCESSFUL) {
		return RTEMS_UNSATISFIED;
	}

	sc =  rtems_timer_initiate_server(
		BSD_TASK_PRIORITY_TIMER,
		BSD_MINIMUM_TASK_STACK_SIZE,
		RTEMS_DEFAULT_ATTRIBUTES
	);
	if (sc != RTEMS_SUCCESSFUL) {
		return RTEMS_UNSATISFIED;
	}

	mutex_init();

	mi_startup();

	return RTEMS_SUCCESSFUL;
}
