/**
 * @file
 *
 * @ingroup rtems_bsd_rtems
 *
 * @brief TODO.
 */

/*
 * Copyright (c) 2009, 2010 embedded brains GmbH.  All rights reserved.
 *
 *  embedded brains GmbH
 *  Obere Lagerstr. 30
 *  82178 Puchheim
 *  Germany
 *  <rtems@embedded-brains.de>
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

/* Necessary to obtain some internal functions */
#define __RTEMS_VIOLATE_KERNEL_VISIBILITY__

#include <rtems/freebsd/machine/rtems-bsd-config.h>

#include <rtems/freebsd/sys/param.h>
#include <rtems/freebsd/sys/types.h>
#include <rtems/freebsd/sys/systm.h>
#include <rtems/freebsd/sys/lock.h>
#include <rtems/freebsd/sys/sx.h>

RTEMS_CHAIN_DEFINE_EMPTY(rtems_bsd_sx_chain);

void
sx_sysinit(void *arg)
{
	struct sx_args *sargs = arg;

	sx_init(sargs->sa_sx, sargs->sa_desc);
}

void
sx_init_flags(struct sx *sx, const char *description, int opts)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;
	rtems_id id = RTEMS_ID_NONE;
	rtems_attribute attr = RTEMS_LOCAL | RTEMS_PRIORITY | RTEMS_BINARY_SEMAPHORE;

	if ((opts & SX_RECURSE) != 0) {
		/* FIXME */
	}

	sc = rtems_semaphore_create(
		rtems_build_name( '_', 'S', 'X', ' '),
		1,
		attr,
		0,
		&id
	);
	BSD_ASSERT_SC(sc);

	sx->lock_object.lo_name = description;
	sx->lock_object.lo_id = id;

	rtems_chain_append(&rtems_bsd_sx_chain, &sx->lock_object.lo_node);
}

void
sx_destroy(struct sx *sx)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	sc = rtems_semaphore_delete( sx->lock_object.lo_id);
	BSD_ASSERT_SC(sc);

	rtems_chain_extract(&sx->lock_object.lo_node);
}

int
_sx_xlock(struct sx *sx, int opts, const char *file, int line)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	BSD_ASSERT((opts & SX_INTERRUPTIBLE) == 0);

	sc = rtems_semaphore_obtain( sx->lock_object.lo_id, RTEMS_WAIT, RTEMS_NO_TIMEOUT);
	BSD_ASSERT_SC(sc);

	return 0;
}

int
_sx_try_xlock(struct sx *sx, const char *file, int line)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	sc = rtems_semaphore_obtain( sx->lock_object.lo_id, RTEMS_NO_WAIT, 0);
	if (sc == RTEMS_SUCCESSFUL) {
		return 1;
	} else if (sc == RTEMS_UNSATISFIED) {
		return 0;
	} else {
		BSD_ASSERT_SC(sc);

		return 0;
	}
}

void
_sx_xunlock(struct sx *sx, const char *file, int line)
{
	rtems_status_code sc = RTEMS_SUCCESSFUL;

	sc = rtems_semaphore_release( sx->lock_object.lo_id);
	BSD_ASSERT_SC(sc);
}

int
_sx_try_upgrade(struct sx *sx, const char *file, int line)
{
	return 1;
}

void
_sx_downgrade(struct sx *sx, const char *file, int line)
{
	/* Do nothing */
}

int
sx_xlocked(struct sx *sx)
{
	Objects_Locations location;
	Semaphore_Control *sema = _Semaphore_Get(sx->lock_object.lo_id, &location);

	if (location == OBJECTS_LOCAL && !_Attributes_Is_counting_semaphore(sema->attribute_set)) {
		int xlocked = sema->Core_control.mutex.holder_id == rtems_task_self();

		_Thread_Enable_dispatch();

		return xlocked;
	} else {
		_Thread_Enable_dispatch();

		BSD_PANIC("unexpected semaphore location or attributes");
	}
}
